"""
Decorator for deploy task
"""


from __future__ import with_statement
import timeit
from fabric.api import task, settings
from py_mina.state import state, set_state
from py_mina.subtasks.deploy import *
from py_mina.echo import echo_task


def deploy_task(on_success=None):
	"""
	Deploy task arguments decorator
	"""

	def deploy_task_wrapper(wrapped_function):
		"""
		Deploy task decorator
		"""

		wrapped_function_name = wrapped_function.__name__


		def _pre_deploy():
			try:
				check_lock()
				lock()
				create_build_path()
				discover_latest_release()
				set_state('pre_deploy', True)
			except Exception as error: 
				set_state('pre_deploy', error)

				raise error # escalate exception to 'deploy_wrapper' function

		def _deploy(*args):
			if state.get('pre_deploy') == True:
				try:
					with cd(fetch('build_to')):
						wrapped_function(*args)
					set_state('deploy', True)
				except Exception as error: 
					set_state('deploy', error)

					raise error # escalate exception to 'deploy_wrapper' function


		def _post_deploy():
			if state.get('deploy') == True:
				try:
					move_build_to_releases()
					link_release_to_current()
					set_state('post_deploy', True)
				except Exception as error: 
					set_state('post_deploy', error)

					raise error # escalate exception to 'deploy_wrapper' function


		def _finalize_deploy():
			if not state.get('success') == True:
				echo_task('Cleaning up failed deploy')

			try:
				cleanup_releases()
				remove_build_path()
				force_unlock()
				set_state('finalize', True)
			except Exception as error:
				set_state('finalize', error)


		def _on_success_deploy():
			if  state.get('success') == True and callable(on_success):
				try: 
					on_success()
					set_state('on_success', True)
				except Exception as error:
					set_state('on_success', error)


		def deploy_wrapper(*args):
			"""
			Runs deploy process on remote server
				1) Pre deploy (lock, build path, latest_release)
				2) Deploy
				3) Post deploy (move to releases, link release to current)
				4) Finalize deploy (cleanup, remove build path, unlock)
				5) Runs `on_success` callback function if deploy successfully finished
				6) Shows deploy stats
			"""
			
			start_time = timeit.default_timer()

			check_deploy_config()

			echo_task('Running "%s" task' % wrapped_function_name)

			with settings(colorize_errors=True):
				try:
					_pre_deploy()
					_deploy(*args)
					_post_deploy()
					set_state('success', True)
				except Exception as error:
					echo_comment(('\n[ERROR]\n%s\n' % error), error=True)
					set_state('success', error)

				_finalize_deploy()
				_on_success_deploy()
				print_deploy_stats(wrapped_function_name, start_time=start_time)

		
		# Copy __name__ and __doc__ from decorated function to decorator function
		deploy_wrapper.__name__ = wrapped_function_name or 'deploy'
		if wrapped_function.__doc__: deploy_wrapper.__doc__ = wrapped_function.__doc__

		# Decorate with `fabric3` task decorator
		return task(deploy_wrapper)

	return deploy_task_wrapper



"""
Git tasks
"""


from __future__ import with_statement
import os
from fabric.api import settings, hide, run, cd
from py_mina.echo import echo_subtask, echo_task
from py_mina.config import fetch, ensure
# from fabric.utils import fastprint

def git_clone():
    """
    Clones repository to tmp build dir
    """

    maybe_clone_git_repository()
    fetch_new_commits()
    use_git_branch()


def maybe_clone_git_repository():
    """
    Clones bare git repository on first deploy
    """

    ensure('repository')
    ensure('scm')

    echo_subtask('Ensuring git repository presence')

    with settings(hide('warnings'), warn_only=True):

        if run('test -d %s' % fetch('scm')).failed:
            echo_subtask("Cloning bare git repository")

            with settings(hide('output')):
                run('git clone {0} {1} --bare'.format(fetch('repository'), fetch('scm')))


def fetch_new_commits():
    """
    Fetches new git commits
    """

    ensure('scm')
    ensure('repository')
    ensure('branch')

    echo_subtask("Fetching new commits")

    with cd(fetch('scm')):
        run('git fetch {0} "{1}:{1}" --force'.format(fetch('repository'), fetch('branch')))

def use_git_branch():
    """
    Clones repository to build dir
    """

    ensure('build_to')
    ensure('scm')
    ensure('branch')

    echo_subtask("Copying code from repository to build folder")

    build_to = fetch('build_to')

    run('git clone {0} {1} --recursive --branch {2}'.format(fetch('scm'), build_to, fetch('branch')))
    run('rm -rf %s' % os.path.join(build_to, '.git'))




"""
Templates
"""

deploy_file_sample = '''"""
py_mina deployfile sample
"""

from py_mina import *
from py_mina.subtasks import (
    git_clone, 
    create_shared_paths, 
    link_shared_paths, 
    rollback_release, 
    force_unlock,
)


# Settings - connection


set('user', 'deploy')
set('hosts', ['localhost'])


# Settings - application


set('verbose', True)
set('deploy_to', '/var/www/example_application')
set('repository', 'https://github.com/py-mina-deploy/py-mina')
set('branch', 'master')


# Settings - shared


set('shared_dirs', [])
set('shared_files', [])


# Tasks


@task
def restart():
    """
    Restarts application on remote server
    """

    run('touch restart.txt')


@deploy_task(on_success=restart)
def deploy():
    """
    Runs deploy process on remote server
    """

    git_clone()
    link_shared_paths()


@setup_task
def setup():
    """
    Runs setup process on remote server
    """

    create_shared_paths()


@task
def rollback():
    """
    Rollbacks to previous release
    """

    rollback_release()


@task
def unlock():
    """
    Forces lockfile removal when previous deploy failed
    """

    force_unlock()
'''



"""
Task decorator (wrapper for `fabric3` @task decorator)
"""


from __future__ import with_statement
import timeit
import fabric.api
from py_mina.echo import echo_task, print_task_stats


def task(wrapped_function):
	"""
	Task function decorator
	"""

	wrapped_function_name = wrapped_function.__name__


	def task_wrapper(*args):
		"""
		Runs task and prints stats at the end
		"""

		echo_task('Running "%s" task\n' % wrapped_function_name)

		start_time = timeit.default_timer()

		with fabric.api.settings(colorize_errors=True):
			try:
				wrapped_function(*args)
			except Exception as e:
				print_task_stats(wrapped_function_name, start_time, e)

				raise e # escalate exception
			else:
				print_task_stats(wrapped_function_name, start_time)


	# Copy __name__ and __doc__ from decorated function to wrapper function
	task_wrapper.__name__ = wrapped_function_name or 'task'
	if wrapped_function.__doc__: task_wrapper.__doc__ = wrapped_function.__doc__

	# Decorate with `fabric3` task decorator
	return fabric.api.task(task_wrapper)



"""
Console output
"""


import timeit
from fabric.colors import *


################################################################################
# Echo
################################################################################


# System


def echo_task(message, error=False):
    color = red if error == True else yellow
    
    print(color('\n=====> %s' % message))


def echo_subtask(message, error=False):
    color = red if error == True else cyan

    print(cyan('\n-----> %s' % message))


# For user


def echo_comment(message, error=False):
    color = red if error == True else blue

    print(color(message))


def echo_status(message, error=False):
    color = red if error == True else green

    print(color(message))


################################################################################
# Print
################################################################################


# Task


def print_task_stats(task_name, start_time, error=None):
    status_tuple = (task_name, time_string(start_time))

    if error != None:
        if (type(error) in [Exception, str]):
            echo_comment(('\n[FATAL ERROR]\n\n%s' % error), error=True)
            
        echo_status(('\n=====> Task "%s" failed %s \n' % status_tuple), error=True)
    else:
        echo_status('\n=====> Task "%s" finished %s \n' % status_tuple)
    

# Helpers


def time_string(start_time):
    stop_time = timeit.default_timer()
    delta_time = stop_time - start_time

    return '(time: %s seconds)' % delta_time



"""
Setup tasks
"""


from __future__ import with_statement
import timeit
import os
import re
from py_mina.config import fetch, ensure, check_config
from fabric.api import run, settings, env, cd, hide
from py_mina.echo import *


################################################################################
# Check
################################################################################


def check_setup_config():
    """
    Checks required config settings for setup task
    """

    check_config(['user', 'hosts', 'deploy_to'])


################################################################################
# Create required and shared 
################################################################################


def create_required_structure():
    """
    Creates required folders (tmp, releases, shared) in `deploy_to` path
    """

    ensure('deploy_to')

    echo_subtask('Creating required structure')

    run('mkdir -p %s' % fetch('deploy_to'))

    for required_path in ['shared', 'releases', 'tmp']:
        run('mkdir -p %s' % os.path.join(fetch('deploy_to'), required_path))


def create_shared_paths():
    """
    Creates shared dirs and touches shared files
    """

    ensure('shared_path')
    ensure('shared_dirs')
    ensure('shared_files')

    echo_subtask('Creating shared paths')

    shared_path = fetch('shared_path')

    # with cd(shared_path):
    for sdir in fetch('shared_dirs'):
        run('mkdir -p %s' % os.path.join(shared_path, sdir))

    for sfile in fetch('shared_files'):
        directory, filename_ = os.path.split(sfile)

        if directory: 
            run('mkdir -p %s' % os.path.join(shared_path, directory))

        filepath = os.path.join(shared_path, sfile)
        
        run('touch ' + filepath)
        run('chmod g+rx,u+rwx ' + filepath)

        recommendation_tuple = (env.host_string, filepath)
        echo_status('\n=====> Don\'t forget to update shared file:\n[%s] %s\n' % recommendation_tuple, error=True)


################################################################################
# SSH known hosts
################################################################################


def add_repo_to_known_hosts():
    """
    Adds repository host to known hosts if possible
    """

    maybe_repository = fetch('repository', default_value='')

    if maybe_repository:
        # Parse repo host
        repo_host_array = re.compile('(@|://)').split(maybe_repository)
        repo_host_part = repo_host_array[-1] if repo_host_array else ''
        repo_host_match = re.compile(':|\/').split(repo_host_part)
        repo_host = repo_host_match[0] if repo_host_match else ''

        # Exit if no host
        if repo_host == '': 
            return

        # Parse port
        repo_port_match = re.search(r':([0-9]+)', maybe_repository)
        repo_port = repo_port_match.group(1) if bool(repo_port_match) else 22

        echo_subtask('Adding repository to known hosts')

        add_to_known_hosts(repo_host, repo_port)


def add_host_to_known_hosts():
    """
    Adds current host to the known hosts

    `fabric3` library sets to `env.host_string` current host where task is executed
    """

    ensure('hosts')

    echo_subtask('Adding current host to known hosts')

    add_to_known_hosts(env.host_string, env.port)


def add_to_known_hosts(host, port):
    cmd = '''
if ! ssh-keygen -H -F {0} &>/dev/null; then
    ssh-keyscan -t rsa -p {1} -H {0} >> ~/.ssh/known_hosts
fi'''.format(host, port)

    with settings(hide('everything'), warn_only=True):
        run(cmd)



"""
Configuration manager
"""


import os
import time
from fabric.api import env
from py_mina.utils import _AttributeDict
import py_mina.state


################################################################################
# Default config
################################################################################


config = _AttributeDict({
    'keep_releases': 7,
    'shared_files': [],
    'shared_dirs': [],
    'abort_on_prompts': False,
    'verbose': False,
    'farbric_config_settings': ['user', 'hosts', 'abort_on_prompts']
})


################################################################################
# Attributes
################################################################################


def ensure(key):
    """
    Ensures presence of config setting
    """

    if not key in config.keys():
        raise Exception('"%s" must be defined' % key)


def fetch(key, default_value=None):
    """
    Gets and ensures config setting or returns `default_value` if provided
    """

    if key in config.keys():
        return config.get(key)
    else:
        if default_value != None:
            return default_value
        else:
            raise Exception('"%s" is not defined' % key)


def set(key, value):
    """
    Sets config setting
    """

    if key in config.get('farbric_config_settings'):
        env.update({ key: value })
        config.update({ key: value })
    elif key == 'deploy_to':
        config.update({
            'deploy_to': value, 
            'scm': os.path.join(value, 'scm'), 
            'shared_path': os.path.join(value, 'shared'), 
            'current_path': os.path.join(value, 'current'), 
            'releases_path': os.path.join(value, 'releases'), 
            'build_to': os.path.join(value, 'tmp', 'build-' + str(time.time())),
        })
    else:
        config.update({ key: value })


# Alias to prevent conflict when importing "py_mina.config" and "py_mina.state" in one file
set_config = set


################################################################################
# Check
################################################################################


def check_config(required_settings=[]):
    """
    Ensures required settings in cofig
    """

    try:
        for setting in required_settings:
            ensure(setting)
    except EnsureConfigError:
        msg = 'Bad config!\nRequired settings: {0}\nCurrent config: {1}'

        raise Exception(msg.format(required_settings, config))  



"""
Deploy tasks
"""

from __future__ import with_statement
import os
from fabric.colors import red, green, yellow
from fabric.api import *
from py_mina.config import fetch, ensure, set, check_config
from py_mina.state import state
from py_mina.echo import *


################################################################################
# Config
################################################################################


def check_deploy_config():
    """
    Check required config settings for deploy task
    """

    check_config(['user', 'hosts', 'deploy_to', 'repository', 'branch'])


################################################################################
# Pre deploy hooks
################################################################################


def check_lock():
    """
    Aborts deploy if lock file is found
    """
    
    echo_subtask("Checking `deploy.lock` file presence")

    with settings(hide('warnings'), warn_only=True):
        if run('test -f %s' % os.path.join(fetch('deploy_to'), 'deploy.lock')).succeeded:
            abort("Another deploy in progress")


def lock():
    """
    Locks deploy
    Parallel deploys are not allowed
    """

    ensure('build_to')

    echo_subtask("Creating `deploy.lock` file")

    run('touch %s' % os.path.join(fetch('deploy_to'), 'deploy.lock'))


def create_build_path():
    """
    Creates tmp dir for building app. 

    Folder will be:
        * deleted -> if build fails
        * moved to releases -> if build succeeds
    """
    
    ensure('build_to')

    echo_subtask("Creating build path")

    with settings(hide('warnings'), warn_only=True):
        build_path = fetch('build_to')

        if run('test -d %s' % build_path).failed:
            run('mkdir -p %s' % build_path)


def discover_latest_release():
    """
    Connects to remote server and discovers next release number
    """

    ensure('releases_path')

    releases_path = fetch('releases_path')

    echo_subtask("Discovering latest release number")

    # Get latest release number
    with settings(hide('warnings'), warn_only=True):
        if run('test -d %s' % releases_path).failed:
            last_release_number = 0
            
            run('mkdir -p %s' % releases_path)
        else:
            releases_respond = run('ls -1p %s | sed "s/\///g"' % releases_path)
            releases_dirs = list(filter(lambda x: len(x) != 0, releases_respond.split('\r\n')))

            if len(releases_dirs) > 0:
                last_release_number = max(map(lambda x: int(x), releases_dirs)) or 0
            else:
                last_release_number = 0

    release_number = last_release_number + 1

    # Set release info to config 
    set('release_number', release_number)
    set('release_to', os.path.join(releases_path, str(release_number)))


################################################################################
# Deploy
################################################################################


def link_shared_paths():
    """
    Links shared paths to build folder
    """

    ensure('build_to')
    ensure('shared_path')
    ensure('shared_dirs')
    ensure('shared_files')

    echo_subtask("Linking shared paths")

    shared = fetch('shared_path')
    build_to = fetch('build_to')

    # Shared dirs
    for sdir in fetch('shared_dirs'):
        relative_path = os.path.join('./', sdir)
        directory, filename_ = os.path.split(relative_path)
        shared_path = os.path.join(shared, sdir)
    
        run('cd %s && mkdir -p %s' % (build_to, directory)) # create parent directory
        run('cd %s && rm -rf %s' % (build_to, relative_path)) # remove directory if it conficts with shared
        run('cd {0} && ln -s "{1}" "{2}"'.format(build_to, shared_path, relative_path)) # link shared to current folder

    # Shared files
    for sfile in fetch('shared_files'):
        shared_path = os.path.join(shared, sfile)
        relative_path = os.path.join('./', sfile)
        directory, filename_ = os.path.split(relative_path)

        run('cd %s && mkdir -p %s' % (build_to, directory)) # create parent directory
        run('cd {0} && ln -sf "{1}" "{2}"'.format(build_to, shared_path, relative_path)) # link shared to current folder


################################################################################
# Post deploy hooks
################################################################################


def move_build_to_releases():
    """
    Moves current build folder to releases path
    """

    ensure('build_to')
    ensure('release_to')

    echo_subtask('Moving from build path to release path')

    run('mv %s %s' % (fetch('build_to'), fetch('release_to')))


def link_release_to_current():
    """
    Creates current release symlink in `deploy_to` path
    """

    ensure('release_to')
    ensure('current_path')

    echo_subtask("Linking release to current")

    release_to = fetch('release_to')

    with cd(release_to):
        run('ln -nfs %s %s' % (release_to, fetch('current_path')))


################################################################################
# Finalize deploy hooks
################################################################################

def cleanup_releases():
    """
    Cleans up old releases
    """

    ensure('releases_path')
    ensure('keep_releases')

    releases_count = str(fetch('keep_releases'))

    echo_subtask("Cleaning up old realeses. Keeping latest %s" % releases_count)

    with cd(fetch('releases_path')), show('debug'):
        cmd = '''
count=$(ls -A1 | sort -rn | wc -l)
remove=$((count > %s ? count - %s : 0))
ls -A1 | sort -rn | tail -n $remove | xargs rm -rf {}'''

        run(cmd % (releases_count, releases_count))


def remove_build_path():
    """
    Removes a temporary build dir
    """

    ensure('build_to')
    
    echo_subtask("Removing build path")

    with settings(hide('stdout', 'warnings'), warn_only=True):
        run('rm -rf %s' % fetch('build_to'))


def force_unlock():
    """
    Forces a deploy unlock
    """

    echo_subtask("Removing `deploy.lock` file")

    run('rm -f %s' % os.path.join(fetch('deploy_to'), 'deploy.lock'))


################################################################################
# Rollback release
################################################################################


def rollback_release():
    """
    Rollbacks latest release
    """
    
    ensure('releases_path')
    ensure('current_path')

    releases_path = fetch('releases_path')

    with(settings(show('debug'))):
        with cd(releases_path):
            echo_subtask('Finding previous release for rollback')
            rollback_release_number = run('ls -1A | sort -n | tail -n 2 | head -n 1')

            if int(rollback_release_number) > 0:
                echo_subtask('Linking previous release to current')
                run('ln -nfs %s/%s %s' % (releases_path, rollback_release_number, fetch('current_path')))

                echo_subtask('Finding latest release')
                current_release = run('ls -1A | sort -n | tail -n 1')

                if int(current_release) > 0:
                    echo_subtask('Removing latest release')
                    run('rm -rf %s/%s' % (releases_path, current_release))
                else:
                    abort('Can\'t find latest release for remove.')
            else:
                abort('Can\'t find previous release for rollback.')


################################################################################
# Print
################################################################################


def get_error_states():
    def find_error_states(x):
        if x == 'success': return False
        
        return state.get(x) not in [True, None]

    return list(filter(find_error_states, state.keys()))


def print_verbose():
    print(yellow('\n=====> Verbose subtasks stats\n'))

    subtask_states = ['pre_deploy', 'deploy', 'post_deploy', 'finalize', 'on_success']

    for subtask_state_name in subtask_states:
        state_value = state.get(subtask_state_name)

        if state_value == None:
            print_value = white('not invoked')
        elif state_value == True:
            print_value = green('success')
        else:
            print_value = red('failed')

        print('%s - %s' % (yellow('* %s' % subtask_state_name), print_value))


def print_deploy_stats(task_name, start_time):
    if fetch('verbose') == True:
        print_verbose()

    error_states = get_error_states()

    if error_states:
        echo_task('Deploy errors', error=True)

        for error_state in error_states:
            echo_comment(('\n[ERROR]\n%s' % state.get(error_state)), error=True)

    if state.get('success'):
        print(yellow('\n-----> Release number: %s' % fetch('release_number')))

    print_stats_args = [task_name, start_time]
    if state.get('success') != True: print_stats_args.append(True)
    print_task_stats(*print_stats_args)
